#include <fcntl.h>
#include <signal.h>
#include <string.h>
#include <unistd.h>

#define errmsg(X) do{ const char* msg = X; write(2,msg,strlen(msg)); }while(0)

#define BUFSIZE 4096
static int fd;
static char* filename;
static char buf[BUFSIZE];

ssize_t do_read(void)
{
  ssize_t rd = read(0, buf, BUFSIZE);
  if(rd == -1)
    rd = 0;
  return rd;
}

ssize_t do_write(ssize_t rd)
{
  ssize_t offset = 0;
  ssize_t wr;
  while(offset < rd) {
    wr = write(fd, buf+offset, rd-offset);
    if(wr == -1)
      break;
    offset += wr;
  }
  return offset;
}

int do_close(void)
{
  if(fd >= 0) {
    if(fsync(fd) == -1) {
      errmsg("filelog: Error syncing the file.\n");
      return 1;
    }
    if(close(fd) == -1) {
      errmsg("filelog: Error closing the file.\n");
      return 1;
    }
  }
  return 0;
}

int do_open(void)
{
  fd = open(filename, O_WRONLY|O_CREAT|O_APPEND, 0666);
  if(fd == -1) {
    errmsg("filelog: Error opening the output file '");
    errmsg(filename);
    errmsg("'\n");
    return 1;
  }
  return 0;
}

int loop(void)
{
  ssize_t rd;
  for(;;) {
    rd = do_read();
    if(!rd)
      return 0;
    if(do_write(rd) < rd)
      return 1;
  }
  return 0;
}

void catch_hup(int flag)
{
  signal(SIGHUP, catch_hup);
  if(do_close() || do_open())
    exit(1);
}

int main(int argc, char* argv[]) 
{
  if(argc < 2) {
    errmsg("usage: filelog filename\n");
    return 1;
  }
  filename = argv[1];
  fd = -1;
  if(do_open())
    return 1;
  signal(SIGHUP, catch_hup);
  if(loop())
    return 1;
  if(do_close())
    return 1;
  return 0;
}
